@@ -29,30 +29,31 @@
 # Calculator is more bother than it is worth... keep exp, ** and /, but drop everything else (use f strings!)
 # Result has SR conflicts but it parses all the test programs
 
-from sly import Lexer, Parser
-import sly
 import logging
 import numpy as np
 from numpy import exp
-import re
 from pathlib import Path
+import re
+from sly import Lexer, Parser
+import sly
 
 logger = logging.getLogger(__name__)
+
 DEBUGFILE = Path.home() / 'aggregate/parser/parser.out'
 
 
 class UnderwritingLexer(Lexer):
     """
-    Implements the Lexer for agg language.
+    Implements the Lexer for the agg language.
 
     """
 
     tokens = {ID, BUILTIN_AGG, BUILTIN_SEV,NOTE,
               SEV, AGG, PORT,
               NUMBER, INFINITY,
-              PLUS, MINUS, TIMES, DIVIDE, HOMOG_MULTIPLY, # SCALE_MULTIPLY, LOCATION_ADD,
-              LOSS, PREMIUM, AT, LR, CLAIMS,
-              XS,
+              PLUS, MINUS, TIMES, DIVIDE, INHOMOG_MULTIPLY,
+              LOSS, PREMIUM, AT, LR, CLAIMS, EXPOSURE, RATE,
+              XS, PICKS,
               DISTORTION,
               CV, WEIGHTS, EQUAL_WEIGHT, XPS,
               MIXED, FREQ, TWEEDIE, ZM, ZT,
@@ -68,10 +69,10 @@ class UnderwritingLexer(Lexer):
     # per manual, need to list longer tokens before shorter ones
     # simple but effective notes
     NOTE = r'note\{[^\}]*\}'  # r'[^\}]+'
-    BUILTIN_AGG = r'agg\.[a-zA-Z][a-zA-Z0-9_:~]*'
-    BUILTIN_SEV = r'sev\.[a-zA-Z][a-zA-Z0-9_:~]*'
+    BUILTIN_AGG = r'agg\.[a-zA-Z][a-zA-Z0-9._:~]*'
+    BUILTIN_SEV = r'sev\.[a-zA-Z][a-zA-Z0-9._:~]*'
     # PORT_BUILTIN = r'port\.[a-zA-Z][a-zA-Z0-9_:~]*'
-    FREQ = 'binomial|pascal|poisson|bernoulli|geometric|fixed' # |empirical'
+    FREQ = 'binomial|pascal|poisson|bernoulli|geometric|fixed|neyman(a|A)?' # |empirical'
     DISTORTION = 'dist(ortion)?'
 
     # number regex including unary minus; need before MINUS else that grabs the minus sign in -3 etc.
@@ -87,18 +88,20 @@ class UnderwritingLexer(Lexer):
     TIMES = r'\*'
     DIVIDE = '/'
     PERCENT = '%'
-    HOMOG_MULTIPLY = '@'
+    INHOMOG_MULTIPLY = '@'
     EQUAL_WEIGHT = '='
     RANGE = ':'
 
     ID['occurrence'] = OCCURRENCE
     ID['unlimited'] = INFINITY
     ID['aggregate'] = AGGREGATE
+    ID['exposure'] = EXPOSURE
     ID['tweedie'] = TWEEDIE
     ID['premium'] = PREMIUM
     ID['tower'] = TOWER
     ID['mixed'] = MIXED
     ID['unlim'] = INFINITY
+    ID['picks'] = PICKS
     ID['prem'] = PREMIUM
     ID['claims'] = CLAIMS
     ID['ceded'] = CEDED
@@ -107,6 +110,7 @@ class UnderwritingLexer(Lexer):
     ID['dsev'] = DSEV
     ID['loss'] = LOSS
     ID['port'] = PORT
+    ID['rate'] = RATE
     ID['net'] = NET
     ID['sev'] = SEV
     ID['agg'] = AGG
@@ -138,7 +142,7 @@ class UnderwritingLexer(Lexer):
         self.lineno += t.value.count('\n')
 
     def error(self, t):
-        print(f"Illegal character '{t.value[0]:s}'")
+        logger.error(f"Illegal character '{t.value[0]:s}'")
         self.index += 1
 
     @staticmethod
@@ -181,20 +185,18 @@ class UnderwritingLexer(Lexer):
 
 class UnderwritingParser(Parser):
     """
-    Implements the Parser for agg language.
+    Implements the Parser for the agg language.
 
     """
 
     # uncomment to write detailed grammar rules
-    debugfile = Path.home() / 'aggregate/parser/parser.out'
+    # debugfile = Path.home() / 'aggregate/parser/parser.out'
+    debugfile = None
     tokens = UnderwritingLexer.tokens
     precedence = (
         # LOW is used to force shift in rules like
-        # expr <- sum or expr <- sum + expr, and empty rules
         ('nonassoc', LOW),
-        # ('nonassoc', LOCATION_ADD),
-        # ('nonassoc', SCALE_MULTIPLY, HOMOG_MULTIPLY),
-        ('nonassoc', HOMOG_MULTIPLY),
+        ('nonassoc', INHOMOG_MULTIPLY),
         ('left', PLUS, MINUS),
         ('left', TIMES, DIVIDE),
         ('right', EXP),
@@ -229,7 +231,7 @@ class UnderwritingParser(Parser):
     @staticmethod
     def enhance_debugfile(f_out=''):
         """
-        Put links in the parser.out debug file, if DEBUGFILE != ''
+        Put links in the parser.out debug file, if DEBUGFILE != ''.
 
         :param f_out: Path or filename of output. If "" then DEBUGFILE.html used.
         :return:
@@ -257,7 +259,7 @@ class UnderwritingParser(Parser):
     @staticmethod
     def _check_vectorizable(value):
         """
-        check the if value can be vectorized
+        Check the value can be vectorized.
 
         """
         if isinstance(value, (float, int, np.ndarray)):
@@ -339,9 +341,9 @@ class UnderwritingParser(Parser):
     @_('AGG name dfreq layers sev_clause occ_reins agg_reins note')
     def agg_out(self, p):
         self.logger(
-            f'agg_out <-- AGG name dfreq dsev occ_reins note', p)
+            f'agg_out <-- AGG name dfreq layers sev_clause occ_reins agg_reins note', p)
         # self.out_dict[("agg", p.name)] =
-        return 'agg', p.name, {'name': p.name, **p.dfreq, **p.sev_clause,
+        return 'agg', p.name, {'name': p.name, **p.dfreq, **p.layers, **p.sev_clause,
                                          **p.occ_reins, **p.agg_reins, 'note': p.note}
 
     @_('AGG name TWEEDIE expr expr expr note')
@@ -375,17 +377,18 @@ class UnderwritingParser(Parser):
         # self.out_dict[('agg', p.name)] = dout
         return 'agg', p.name, dout
 
-    @_('AGG name builtin_agg note')
+    @_('AGG name builtin_agg agg_reins note')
     def agg_out(self, p):
         # for use when you change the agg and/or  want a new name
         self.logger(
             f'agg_out <-- AGG name builtin_aggregate note', p)
-        # self.out_dict[("agg", p.name)] =
-        return 'agg', p.name, {'name': p.name, **p.builtin_agg, 'note': p.note}
+        # rename; NOTE!! the code below will overwrite the new name!
+        del p.builtin_agg['name']
+        return 'agg', p.name, {'name': p.name, **p.builtin_agg, **p.agg_reins, 'note': p.note}
 
     @_('builtin_agg agg_reins note')
     def agg_out(self, p):
-        # no change to the built in agg, allows agg.A as a legitimate agg (called A)
+        # no change to the builtin agg, allows agg.A as a legitimate agg (called A)
         self.logger(
             f'agg_out <-- builtin_agg agg_reins note', p)
         # print(p.builtin_agg)
@@ -460,9 +463,9 @@ class UnderwritingParser(Parser):
     def freq(self, p):
         self.logger(
             f'freq <-- FREQ expr {p.FREQ}', p)
-        if p.FREQ != 'binomial':
+        if p.FREQ not in ['binomial', 'neyman', 'neymana', 'neymanA']:
             logger.warning(
-                f'Illogical choice of frequency {p.FREQ}, expected binomial')
+                f'Illogical choice of frequency {p.FREQ}, expected binomial or neyman A')
         return {'freq_name': p.FREQ, 'freq_a': p.expr}
 
     @_('FREQ')
@@ -525,34 +528,27 @@ class UnderwritingParser(Parser):
             f'reins_clause <-- tower', p)
         limit = p.tower[0]
         attach = p.tower[1]
-        return [(1, l, a) for l, a in zip(limit, attach)]
+        return [(1.0, l, a) for l, a in zip(limit, attach)]
 
     @_('expr XS expr')
     def reins_clause(self, p):
         self.logger(
             f'reins_clause <-- expr XS expr {p[0]} xs {p[2]}', p)
-        if np.isinf(p[0]):
-            return (1, p[0], p[2])
-        else:
-            # y p/o y xs a
-            return (p[0], p[0], p[2])
+        return (1.0, p[0], p[2])
 
     @_('expr SHARE_OF expr XS expr')
     def reins_clause(self, p):
         self.logger(
             f'reins_clause <-- expr SHARE_OF expr XS expr {p[0]} s/o {p[2]} xs {p[4]}', p)
-        # here expr is the proportion...
-        # TODO should always store as a proportion; that holds for np.inf
-        if np.isinf(p[2]):
-            return (p[0], p[2], p[4])
-        else:
-            return (p[0] * p[2], p[2], p[4])
+        # here expr is the proportion...always store as a proportion
+        return (p[0], p[2], p[4])
 
     @_('expr PART_OF expr XS expr')
     def reins_clause(self, p):
         self.logger(
             f'reins_clause <-- expr PART_OF expr XS expr {p[0]} p/o {p[2]} xs {p[4]}', p)
-        return (p[0], p[2], p[4])
+        # here expr is the currency amount of cover
+        return (p[0] / p[2], p[2], p[4])
 
     # severity term ============================================
     @_('SEV sev %prec LOW')
@@ -572,6 +568,11 @@ class UnderwritingParser(Parser):
             del built_in_dict['name']
         return built_in_dict
 
+    @_('sev picks')
+    def sev(self, p):
+        self.logger(f'sev <-- sev picks', p)
+        return {**p.sev, **p.picks}
+
     @_('sev "!"')
     def sev(self, p):
         self.logger(f'sev <-- unconditional (conditional=False) flag set', p)
@@ -584,33 +585,33 @@ class UnderwritingParser(Parser):
         p.sev['sev_loc'] = UnderwritingParser._check_vectorizable(
             p.sev.get('sev_loc', 0))
         sign = 1 if p[1]=='+' else -1
-        p.numbers = UnderwritingParser._check_vectorizable(p.numbers)
-        p.sev['sev_loc'] += sign * p.numbers
+        p_numbers = UnderwritingParser._check_vectorizable(p.numbers)
+        p.sev['sev_loc'] += sign * p_numbers
         return p.sev
 
     @_('numbers TIMES sev')
     def sev(self, p):
         self.logger(f'sev <-- numbers TIMES sev', p)
-        p.numbers = UnderwritingParser._check_vectorizable(p.numbers)
+        p_numbers = UnderwritingParser._check_vectorizable(p.numbers)
         if 'sev_mean' in p.sev:
             p.sev['sev_mean'] = UnderwritingParser._check_vectorizable(
                 p.sev.get('sev_mean', 0))
-            p.sev['sev_mean'] *= p.numbers
+            p.sev['sev_mean'] *= p_numbers
         # only scale if there is a scale (otherwise you double count)
         if 'sev_scale' in p.sev:
             p.sev['sev_scale'] = UnderwritingParser._check_vectorizable(
                 p.sev.get('sev_scale', 0))
-            p.sev['sev_scale'] *= p.numbers
+            p.sev['sev_scale'] *= p_numbers
         if 'sev_mean' not in p.sev:
             # e.g. Pareto has no mean and it is important to set the scale
             # but if there is a mean it handles the scaling and setting scale will
             # confuse the distribution maker
-            p.sev['sev_scale'] = p.numbers
+            p.sev['sev_scale'] = p_numbers
         # if there is a location it needs to scale too --- that's a curious choice!
         if 'sev_loc' in p.sev:
             p.sev['sev_loc'] = UnderwritingParser._check_vectorizable(
                 p.sev['sev_loc'])
-            p.sev['sev_loc'] *= p.numbers
+            p.sev['sev_loc'] *= p_numbers
         # logger.error(str(p.sev))
         return p.sev
 
@@ -631,7 +632,7 @@ class UnderwritingParser(Parser):
     def sev(self, p):
         self.logger(
             f'sev <-- ids numbers weights', p)
-        return {'sev_name': p.ids, 'sev_a':  p[1], 'sev_wt': p.weights}
+        return {'sev_name': p.ids, 'sev_a':  p[1], 'sev_scale': 1.0,  'sev_wt': p.weights}
 
     # no weights with xps terms
     @_('ids xps')
@@ -641,7 +642,7 @@ class UnderwritingParser(Parser):
 
     @_('ids')
     def sev(self, p):
-        # for norm expon uniform, zero parameter severities
+        # for norm expon uniform levy, zero parameter severities
         # need to make sure there is a scale
         self.logger(
             f'sev <-- ids, zero parameter severity {p.ids}', p)
@@ -676,6 +677,12 @@ class UnderwritingParser(Parser):
             b = p.dprobs
         return {'freq_name': 'empirical', 'freq_a': p.doutcomes, 'freq_b': b, 'exp_en': -1}
 
+    @_('PICKS "[" numberl "]" "[" numberl "]"')
+    def picks(self, p):
+        self.logger('picks <-- PICKS "[" numberl "]" "[" numberl "]"', p)
+
+        return {'sev_pick_attachments': p[2], 'sev_pick_losses': p[5]}
+
     # never valid for this to be a single number not in [], using this
     # format rather than numbers enforces an actual list
     @_('"[" numberl "]"')
@@ -746,8 +753,12 @@ class UnderwritingParser(Parser):
         # doutcomes allows a list, range, or range with step
         self.logger(f'tower <-- tower doutcomes', p)
         breaks = p.doutcomes
-        if breaks[0] != 0:
-            breaks = np.hstack((0., breaks))
+        # do not want this. it means net == 0 and ceded== gross in total which
+        # is rarely what you want. User can put in themselves.
+        # if breaks[0] != 0:
+        #     breaks = np.hstack((0., breaks))
+        # if not np.isinf(breaks[-1]):
+        #     breaks = np.hstack((breaks, np.inf))
         limits = np.diff(breaks)
         attach = breaks[:-1]
         # logger.info('\n'.join([f'{x} xs {y}' for x, y in zip(limits, attach)]))
@@ -781,6 +792,11 @@ class UnderwritingParser(Parser):
             f'exposures <-- numbers PREMIUM AT numbers LR', p)
         return {'exp_premium': p[0], 'exp_lr': p[3], 'exp_el': np.array(p[0]) * np.array(p[3])}
 
+    @_('numbers EXPOSURE AT numbers RATE')
+    def exposures(self, p):
+        self.logger(f'exposures <-- numbers EXPOSURE AT numbers RATE', p)
+        return {'exp_premium': p[0], 'exp_lr': p[3], 'exp_el': np.array(p[0]) * np.array(p[3])}
+
     # ID =======================================================
     @_('"[" idl "]"')
     def ids(self, p):
@@ -806,16 +822,17 @@ class UnderwritingParser(Parser):
         return p.ID
 
     # elements made from named portfolios ========================
-    @_('expr HOMOG_MULTIPLY builtin_agg')
+    @_('expr INHOMOG_MULTIPLY builtin_agg')
     def builtin_agg(self, p):
         """  inhomogeneous change of scale """
         self.logger(
-            f'builtin_agg <-- expr HOMOG_MULTIPLY builtin_agg', p)
+            f'builtin_agg <-- expr INHOMOG_MULTIPLY builtin_agg', p)
         bid = p.builtin_agg.copy()
         bid['name'] += '_i_scaled'
-        bid['exp_en'] = bid.get('exp_en', 0) * p.expr
-        bid['exp_el'] = bid.get('exp_el', 0) * p.expr
-        bid['exp_premium'] = bid.get('exp_premium', 0) * p.expr
+
+        bid['exp_en'] = self._check_vectorizable(bid.get('exp_en', 0)) * p.expr
+        bid['exp_el'] = self._check_vectorizable(bid.get('exp_el', 0)) * p.expr
+        bid['exp_premium'] = self._check_vectorizable(bid.get('exp_premium', 0)) * p.expr
         return bid
 
     @_('expr TIMES builtin_agg')
@@ -826,15 +843,15 @@ class UnderwritingParser(Parser):
         bid = p.builtin_agg
         bid['name'] += '_homog_scaled'
         if 'sev_mean' in bid:
-            bid['sev_mean'] = bid['sev_mean'] * p.expr
+            bid['sev_mean'] = self._check_vectorizable(bid['sev_mean']) * p.expr
         if 'sev_scale' in bid:
-            bid['sev_scale'] = bid['sev_scale'] * p.expr
+            bid['sev_scale'] = self._check_vectorizable(bid['sev_scale']) * p.expr
         if 'sev_loc' in bid:
-            bid['sev_loc'] = bid['sev_loc'] * p.expr
-        bid['exp_attachment'] = bid.get('exp_attachment', 0) * p.expr
-        bid['exp_limit'] = bid.get('exp_limit', np.inf) * p.expr
-        bid['exp_el'] = bid.get('exp_el', 0) * p.expr
-        bid['exp_premium'] = bid.get('exp_premium', 0) * p.expr
+            bid['sev_loc'] = self._check_vectorizable(bid['sev_loc']) * p.expr
+        bid['exp_attachment'] = self._check_vectorizable(bid.get('exp_attachment', 0)) * p.expr
+        bid['exp_limit'] = self._check_vectorizable(bid.get('exp_limit', np.inf)) * p.expr
+        bid['exp_el'] = self._check_vectorizable(bid.get('exp_el', 0)) * p.expr
+        bid['exp_premium'] = self._check_vectorizable(bid.get('exp_premium', 0)) * p.expr
         return bid
 
     @_('builtin_agg PLUS expr', 'builtin_agg MINUS expr')
@@ -1020,29 +1037,22 @@ def grammar(add_to_doc=False, save_to_fn=''):
     """
     Write the grammar at the top of the file as a docstring
 
-    To work with multi rules use ', ' all on one line
+    To work with multi-rules enter them on one line, like so::
+
+        @_('builtin_agg PLUS expr', 'builtin_agg MINUS expr')
+
+    :param add_to_doc: add the grammar to the docstring
+    :param save_to_fn: save the grammar to a file
     """
 
-    start_string = '''Language Specification
-======================
+    pout = Path(__file__).parent / '../doc/4_agg_language_reference/ref_include.rst'
 
-The ```agg``` Language Grammar:
-
-::
-
-'''
-    end_string = '''
-References
-----------
-'''
-    pout = Path(__file__).parent / '../doc/5_Langauge_Reference.rst'
-    pin = Path(__file__)
-
-    txt = pin.read_text(encoding='utf-8')
+    # get the grammar from the top of the file
+    txt = Path(__file__).read_text(encoding='utf-8')
     stxt = txt.split('@_')
     ans = {}
-    # 3:-2 get rid of junk at top and bottom (could change if file changes)
-    for it in stxt[3:-2]:
+    # 3:-3 get rid of junk at top and bottom (could change if file changes)
+    for it in stxt[3:-3]:
         if it.find('# def') >= 0:
             # skip rows with a comment between @_ and def
             pass
@@ -1057,7 +1067,7 @@ References
             try:
                 b1 = b[1].split("(self, p):")[0].strip()
             except:
-                print(it)
+                logger.warning(f'Unexpected multirule behavior {it}')
                 exit()
             if b1 in ans:
                 ans[b1] += b0
@@ -1105,7 +1115,7 @@ EXP                     ::= 'exp'
 
 EXPONENT                ::= '^|**'
 
-HOMOG_MULTIPLY          ::= "@"
+INHOMOG_MULTIPLY        ::= "@"
 
 INFINITY                ::= 'inf|unlim|unlimited'
 
@@ -1139,24 +1149,16 @@ WEIGHTS                 ::= 'wts|wt'
 
 XPS                     ::= 'xps'
 
-xs                      ::= "xs|x"
+XS                      ::= "xs|x"
 
 '''
 
-# PLUS =                  ::= '+'
-# MINUS                   ::= '-'
-# TIMES                   ::= '*'
-# DIVIDE                  ::=  '/'
-
     s += lang_words
-    # actually put into this file (uncomment)
+    # create for docs in one file (that gets included by rst)
     if add_to_doc is True:
-        txt = pout.read_text(encoding='utf-8')
-        st = txt.find(start_string) + len(start_string)
-        end = txt.find(end_string)
-        txt = txt[0:st] + s + txt[end:]
-        pout.write_text(txt, encoding='utf-8')
+        pout.write_text(s, encoding='utf-8')
 
+    # save to user folder gammar
     if save_to_fn == '':
         save_to_fn = Path.home() / 'aggregate/parser/grammar.md'
     Path(save_to_fn).write_text(s, encoding='utf-8')
